/**
 * constructor
 * @param {function} connected_cb: connected callback function.
 * @param {function} disconnected_cb: disconnected callback function.
 * @param {function} message_cb: message callback function. this function is called when receive vad packet.
 * @param {function} epd_cb: epd state changed callback. this function is called when epd on / off packet received.
 */
function WsClient(connected_cb, disconnected_cb, message_cb, epd_cb)
{
  this.connected = false;
  this.ws = null;
  this.connected_cb = connected_cb;
  this.disconnected_cb = disconnected_cb;
  this.message_cb = message_cb;
  this.epd_cb = epd_cb;

  this.EPD_STATE_CHANGED_DELAY = 1000;
}


/**
 * connect to ws server
 */
WsClient.prototype.connect = function(ip, port)
{
  console.log('connect to client ' + ip + ':' + port);

  // create web socket
  this.ws = new WebSocket("ws://" + ip + ":" + port + "/websocket");

  // on open function
  this.ws.onopen = this.onopen.bind(this);

  // on close function
  this.ws.onclose = this.onclose.bind(this);

  // on message function
  this.ws.onmessage = this.onmessage.bind(this);

  // on error function
  this.ws.onerror = this.onerror.bind(this);
};


/**
 * open connection
 */
WsClient.prototype.onopen = function()
{
  console.log('websocket client is opened.');
  this.connected = true;
};


/**
 * on close
 * @param  {[type]} e: what
 */
WsClient.prototype.onclose = function(e)
{
  console.log('websocket client is closed.');
  this.connected = false;
  this.disconnected_cb();
};


/**
 * [onerror description]
 * @param  {[type]} e
 * @return {[type]}
 */
WsClient.prototype.onerror = function(e)
{
  console.log('on error');
};


/**
 * on message
 * @param  {[type]} e
 */
WsClient.prototype.onmessage = function(e)
{
  if(Object.prototype.toString.call(e.data).slice(8, -1) == "Blob"){
    // add onload function
    var reader = new FileReader();
    reader.onload = this.onload.bind(this);
    reader.readAsArrayBuffer(e.data);

  } else {
    console.log("%%%%% text data received : " +e.data);
    if(e.data == "[VAD SERVER Connected]"){
      this.connected_cb();

    } else if (e.data == "[VAD SERVER Disconnected]"){
      this.disconnected_cb();

    } else if (e.data.indexOf("[SS] epdoff") >= 0 ){
      var me = this;
      setTimeout(function() { me.epd_cb(false);}, this.EPD_STATE_CHANGED_DELAY);

    } else if (e.data.indexOf("[SS] epdon") >= 0 ){
      var me = this;
      setTimeout(function() { me.epd_cb(true);}, this.EPD_STATE_CHANGED_DELAY);

    } else {
      console.log('unknown frame');
    }
  }
};


/**
 * onload function
 * @param  {[type]} e
 */
WsClient.prototype.onload = function(e)
{
  var data = new Uint8Array(e.target.result);
  var head = data[0];

  if(head == VAD_HEADER){
    var recvData = this.parseFrame( data );
    this.message_cb( recvData.type, recvData.frame );
  }

  data = null;
};


/**
 * disconnect to server
 */
WsClient.prototype.disconnect = function()
{
  if ( this.isConnected() ) {
    this.ws.close();
    delete this.ws;
    this.ws = null;
    this.connected = false;
  }
};


/**
 * check connection
 * @return {Boolean} true for connected / false for disconnected.
 */
WsClient.prototype.isConnected = function()
{
  return this.connected;
};


/**
 * parse vad frame
 * @param  {Uint8Array} data: data frame
 * @return {number} type: vad type
 * @return {Uint8Array}: vad data frame
 */
WsClient.prototype.parseFrame = function(data)
{
  var receivedType;
  var type_offset = WEBSOCKET_HEADER_SIZE + VAD_FRAME_HEADER_SIZE;
  receivedType = data[type_offset - 1];
  frame = data.subarray(type_offset);
  return {
    type: receivedType,
    frame: frame
  };
};


/**
 * send string to ws server
 * @param  {string} _data: send string
 */
WsClient.prototype.sendStr = function(_data)
{
  if ( this.isConnected() ) {
    this.ws.send(_data);
  }
};


/**
 * send binary data
 * @param  {Uint8Array} _data: send data
 */
WsClient.prototype.sendBin = function(_data)
{
  if ( this.isConnected() ) {
    this.ws.send(_data);
  }
};

